#!/usr/bin/env python

# vim: set expandtab shiftwidth=4 tabstop=4:

#
# Jacek Fedorynski <jfedor@jfedor.org>
# http://www.jfedor.org/
#

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
import pygame, sys, math, random, ctypes
#import psyco

#psyco.full()

width = 1024
height = 768
fullscreen = 0

block_counter = 0

class Block:
    def __init__(self, pos, angle=0, tmp=False):
        self.pos = pos
        self.tmp = True
        self.physical = False
        self.selected = False
        self.setAngle(angle)
        if not self.tmp:
            self.untmp()

    def untmp(self):
        global block_counter
        block_counter += 1
        self.block_id = block_counter
        self.tmp = False

    def make_physical(self):
        physics.create_block(self.block_id, ctypes.c_float(self.pos[0]), ctypes.c_float(self.pos[2]), ctypes.c_float(self.angle))
        self.rotmatrix = Float16()
        self.physical = True

    def setAngle(self, newangle=None):
        if not (newangle is None):
            self.angle = newangle
        x,y,z = self.getPosition()
        ct = math.cos(self.angle)
        st = math.sin(self.angle)
        self.rotmatrix = [ct, 0., -st, 0, 0., 1., 0., 0, st, 0., ct, 0, x, y, z, 1]

    def adjustAngle(self, nextangle):
        self.setAngle((self.angle + nextangle)/2.0)

    def getPosition(self):
        global Float16
        if not self.physical:
            return self.pos
        else:
            raise Exception('not reached')
            m = Float16()
            physics.get_block_matrix(self.block_id, m)
            return (m[12], m[14])

    def draw(self):
        global block_display_list
        global draw_flag
        if not self.tmp:
            glPushName(self.block_id)
        glPushMatrix()
        if self.physical:
            physics.get_block_matrix(self.block_id, self.rotmatrix)
        else:
            glMultMatrixf(self.rotmatrix)
        if self.tmp:
            glColor3d(0.0, 1.0, 0.0)
            glScaled(1, 2, 0.4)
            glDisable(GL_LIGHTING)
            glutWireCube(1.0)
            glEnable(GL_LIGHTING)
        else:
            if self.selected:
                glColor3d(1.0, 0.0, 1.0)
            glCallList(block_display_list)
            if self.selected:
                glColor3d(1.0, 0.0, 0.0)
        glPopMatrix()
        if not self.tmp:
            glPopName()

def draw_floor():
    glPushName(0)
    glColor3d(0.0, 0.5, 0.0)
    glBegin(GL_QUADS)
    glNormal3d(0, 1, 0)
    glVertex3d(-500, 0, 500)
    glVertex3d(500, 0, 500)
    glVertex3d(500, 0, -500)
    glVertex3d(-500, 0, -500)
    glEnd()
    glPopName()

def setup_projection(center, distance, yaw, pitch, selection = False, x = None, y = None):
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    if selection:
        viewport = glGetIntegerv(GL_VIEWPORT)
        gluPickMatrix(x, viewport[3] - y, 1, 1, viewport)
    gluPerspective(30, 1.0*width/height, 0.01, 1000.0)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    glEnable(GL_DEPTH_TEST)
    gluLookAt(0, 0, distance, 0, 0, 0, 0.0, 1.0, 0.0)
    glRotated(pitch, 1, 0, 0)
    glRotated(yaw, 0, 1, 0)
    glTranslated(-center[0], -center[1], -center[2])

def do_selection(x, y, floor_only = False):
    global blocks, center, distance, yaw, pitch
    glSelectBuffer(1024)
    glRenderMode(GL_SELECT)
    glInitNames()
    setup_projection(center, distance, yaw, pitch, True, x, y)
    draw_floor()
    if not floor_only:
        for b in blocks.values():
            b.draw()
    glFlush()
    names = glRenderMode(GL_RENDER)
    return names

def create_blocks_from_tmp():
    global tmp_blocks, selected_block, chained_block
    if tmp_blocks and selected_block and chained_block:
        selected_block.adjustAngle(tmp_blocks[0].angle)
    for b in tmp_blocks:
        b.untmp()
        blocks[b.block_id] = b
        if selected_block:
            selected_block.selected = False
        selected_block = b
        selected_block.selected = True
    tmp_blocks = []

physics = ctypes.CDLL('./physics.so')
Float16 = ctypes.c_float * 16

pygame.init()
glutInit([])
pygame.display.gl_set_attribute(pygame.GL_DEPTH_SIZE, 24)
screen = pygame.display.set_mode((width, height), pygame.OPENGL | pygame.DOUBLEBUF | (fullscreen * pygame.FULLSCREEN))
pygame.display.set_caption('Domino')
myfont = pygame.font.Font("VeraMono.ttf", 24)
myfont2 = pygame.font.Font("VeraMono.ttf", 48)
overlay = pygame.Surface((width, height)).convert_alpha()
clock = pygame.time.Clock()

glEnable(GL_LIGHTING)
glEnable(GL_COLOR_MATERIAL)
glEnable(GL_NORMALIZE)
glShadeModel(GL_SMOOTH)
glClearColor(0.7, 0.7, 0.7, 0.0)
glViewport(0, 0, width, height)

glLightfv(GL_LIGHT0,GL_POSITION,[30,30,30,0])
glLightfv(GL_LIGHT0,GL_DIFFUSE,[1,1,1,1])
glLightfv(GL_LIGHT0,GL_SPECULAR,[1,1,1,1])
glEnable(GL_LIGHT0)

block_display_list = glGenLists(1)
glEnable(GL_DEPTH_TEST)
glNewList(block_display_list, GL_COMPILE)
#glColor3d(1.0, 0.0, 0.0)
glScaled(1, 2, 0.4)
glutSolidCube(1)
glEndList()

hud_texture = glGenTextures(1)
clear_screen = '\0' * 2048*2048*2

blocks = {}
tmp_blocks = []
selected_block = None
chained_block = False

stage = 0
mode = 'none'
update_hud = True

center = [0, 0, 0]
yaw = -45
pitch = 45
distance = 50

stepping = 1.2
draw_flag = True

done = False
while not done:
    for event in pygame.event.get():
        key_mods = pygame.key.get_mods()
        if event.type == pygame.QUIT:
            sys.exit()
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                done = True
            elif event.key == pygame.K_SPACE:
                draw_flag = not draw_flag
            elif event.key == pygame.K_RETURN:
                if stage == 0:
                    tmp_blocks = []
                    prev_time = pygame.time.get_ticks()
                    physics.init()
                    for b in blocks.values():
                        b.make_physical()
                    physics.apply_force(blocks[1].block_id)
                    stage = 10
            elif event.key == pygame.K_LEFT:
                yaw = (yaw + 10) % 360
            elif event.key == pygame.K_RIGHT:
                yaw = (yaw - 10) % 360
        elif event.type == pygame.MOUSEMOTION:
            if mode in ['none', 'create'] and stage < 10:
                if key_mods & pygame.KMOD_SHIFT:
                    if selected_block:
                        tmp_blocks = []
                        chained_block = True
                        names = do_selection(event.pos[0], event.pos[1], True)
                        pos1 = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)
                        pos0 = selected_block.getPosition()
                        dist = math.sqrt((pos1[0]-pos0[0])**2+(pos1[2]-pos0[2])**2)
                        dist0 = dist/stepping
                        angle = math.pi/2+math.atan2(pos1[2]-pos0[2], pos1[0]-pos0[0])
                        for i in range(int(math.floor(dist0))):
                            tmp_blocks.append(Block([pos0[0]+((i+1)/dist0)*(pos1[0]-pos0[0]), 1, pos0[2]+((i+1)/dist0)*(pos1[2]-pos0[2])], -angle, True))
                    if mode == 'create':
                        create_blocks_from_tmp()
                else:
                    tmp_blocks = []
                    chained_block = False
                    names = do_selection(event.pos[0], event.pos[1], True) # FIXME
                    if len(names) == 1 and names[0].names[0] == 0:
                        pos1 = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)
                        if selected_block:
                            my_angle = selected_block.angle
                        else:
                            my_angle = 0
                        b = Block([pos1[0], 1, pos1[2]], my_angle, True)
                        tmp_blocks.append(b)
            elif mode == 'pan':
                names = do_selection(event.pos[0], event.pos[1], True)
                pos1 = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)
                center[0] -= pos1[0] - drag_start_pos[0]
                center[2] -= pos1[2] - drag_start_pos[2]
            elif mode == 'rotateview':
                names = do_selection(event.pos[0], event.pos[1], True)
                pos1 = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)
                v1 = [drag_start_pos[0] - center[0], drag_start_pos[2] - center[2]]
                v2 = [pos1[0] - center[0], pos1[2] - center[2]]
                rot_angle = math.atan2(v2[1],v2[0]) - math.atan2(v1[1],v1[0])
                yaw -= math.degrees(rot_angle)
            elif mode == 'rotate':
                pos0 = selected_block.getPosition()
                names = do_selection(event.pos[0], event.pos[1], True)
                pos1 = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)
                v1 = [drag_start_pos[0] - pos0[0], drag_start_pos[2] - pos0[2]]
                v2 = [pos1[0] - pos0[0], pos1[2] - pos0[2]]
                selected_block.setAngle(selected_block.angle - (math.atan2(v2[1],v2[0]) - math.atan2(v1[1],v1[0])))
                drag_start_pos = pos1
            elif mode == 'move':
                names = do_selection(event.pos[0], event.pos[1], True)
                pos1 = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)
                selected_block.pos[0] += pos1[0] - drag_start_pos[0]
                selected_block.pos[2] += pos1[2] - drag_start_pos[2]
                selected_block.setAngle()
                drag_start_pos = pos1
            elif mode == 'delete':
                names = do_selection(event.pos[0], event.pos[1], False)
                names.sort(key=lambda x: x.near)
                if names[0].names[0] != 0:
                    if selected_block and selected_block is blocks[names[0].names[0]]:
                        selected_block = None
                    del blocks[names[0].names[0]]
        elif event.type == pygame.MOUSEBUTTONUP:
            mode = 'none'
        elif event.type == pygame.MOUSEBUTTONDOWN:
            if event.button == 4:
                distance *= 1.1
            elif event.button == 5:
                distance *= 0.91
                if distance < 8:
                    distance = 8
            elif event.button == 3:
                names = do_selection(event.pos[0], event.pos[1], True)
                drag_start_pos = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)
                if key_mods & pygame.KMOD_CTRL:
                    mode = 'rotateview'
                else:
                    mode = 'pan'
            elif event.button == 1:
                if key_mods & pygame.KMOD_ALT:
                    mode = 'delete'
                    tmp_blocks = []
                    names = do_selection(event.pos[0], event.pos[1], False)
                    names.sort(key=lambda x: x.near)
                    if names[0].names[0] != 0:
                        if selected_block and selected_block is blocks[names[0].names[0]]:
                            selected_block = None
                        del blocks[names[0].names[0]]
                elif key_mods & pygame.KMOD_CTRL:
                    if selected_block:
                        mode = 'rotate'
                        names = do_selection(event.pos[0], event.pos[1], True)
                        drag_start_pos = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)
                else:
                    mode = 'create'
                    names = do_selection(event.pos[0], event.pos[1], False)
                    names.sort(key=lambda x: x.near)
                    if names[0].names[0] == 0:
                        create_blocks_from_tmp()
                    else:
                        if selected_block:
                            selected_block.selected = False
                        selected_block = blocks[names[0].names[0]]
                        selected_block.selected = True
                        mode = 'move'
                        names = do_selection(event.pos[0], event.pos[1], True)
                        drag_start_pos = gluUnProject(event.pos[0], height-event.pos[1], (names[0].near+names[0].far)/2)

    if stage == 10:
        time_now = pygame.time.get_ticks()
        time_delta = time_now - prev_time
        prev_time = time_now
        physics.step(ctypes.c_float(time_delta/1000.0))
        #print pygame.time.get_ticks() - time_now

    setup_projection(center, distance, yaw, pitch)

    if draw_flag:
        draw_floor()

        glColor3d(1.0, 0.0, 0.0)
        for b in blocks.values() + tmp_blocks:
            b.draw()

    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    gluOrtho2D(0, width, height, 0)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    glDisable(GL_LIGHTING)
    glDisable(GL_DEPTH_TEST)

    glBindTexture(GL_TEXTURE_2D, hud_texture)
    if update_hud:
        glPixelStorei(GL_PACK_ALIGNMENT, 1)
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 2048, 2048, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, clear_screen)
        txt = myfont.render("quick brown fox jumps over the lazy dog", 1, (255, 0, 0))
        #glTexSubImage2D(GL_TEXTURE_2D, 0, int(width/2.0-txt.get_width()/2.0), height-72, txt.get_width(), txt.get_height(), GL_RGBA, GL_UNSIGNED_BYTE, pygame.image.tostring(txt, "RGBA"))
        update_hud = False

    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    glEnable(GL_TEXTURE_2D)
    glColor3d(1.0, 1.0, 1.0)

    glBegin(GL_QUADS)
    glTexCoord2d(0, 0)
    glVertex2d(0, 0)
    glTexCoord2d(0, 1)
    glVertex2d(0, 2048)
    glTexCoord2d(1, 1)
    glVertex2d(2048, 2048)
    glTexCoord2d(1, 0)
    glVertex2d(2048, 0)
    glEnd()

    glDisable(GL_BLEND)
    glEnable(GL_LIGHTING)
    glDisable(GL_TEXTURE_2D)

    glDisable(GL_LIGHTING)
    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    txt = myfont.render(str(len(blocks)) + " %.1f" % clock.get_fps(), 1, (255, 0, 0))
    pixeldata = pygame.image.tostring(txt, "RGBA", 1)
    glRasterPos2d(width-txt.get_width(), txt.get_height())
    glDrawPixels(txt.get_width(), txt.get_height(), GL_RGBA, GL_UNSIGNED_BYTE, pixeldata)
    glDisable(GL_BLEND)
    glEnable(GL_LIGHTING)

    pygame.display.flip()
    clock.tick(100)
