#include <map>
#include <btBulletDynamicsCommon.h>
#include <GL/gl.h>

extern "C" {

std::map<int, btRigidBody*> blocks;

btDynamicsWorld* world;

int init() {
	btDefaultCollisionConfiguration* collisionConfiguration = new btDefaultCollisionConfiguration();
	btCollisionDispatcher* dispatcher = new btCollisionDispatcher(collisionConfiguration);
	//dispatcher->registerCollisionCreateFunc(BOX_SHAPE_PROXYTYPE,BOX_SHAPE_PROXYTYPE,new BoxBoxCollisionAlgorithm::CreateFunc);
	btSequentialImpulseConstraintSolver* solver = new btSequentialImpulseConstraintSolver;
	btVector3 worldAabbMin(-1000,-1000,-1000);
        btVector3 worldAabbMax(1000,1000,1000);
        btBroadphaseInterface* broadphase = new bt32BitAxisSweep3(worldAabbMin,worldAabbMax);
	world = new btDiscreteDynamicsWorld(dispatcher,broadphase,solver,collisionConfiguration);
	world->setGravity(btVector3(0, -30, 0));

	btCollisionShape* shape = new btStaticPlaneShape(btVector3(0.0,1.0,0.0),0.0);
	btTransform trans;
	trans.setIdentity();
	btDefaultMotionState* myMotionState = new btDefaultMotionState(trans);
	world->addRigidBody(new btRigidBody(0.0, myMotionState, shape));

	return 0;
}

int step(float delta) {
	return world->stepSimulation(delta, 3, 1./60.);
}

int create_block(int id, float x, float z, float angle) {
	btCollisionShape* shape;
	btTransform trans;
	btVector3 localInertia;
	float mass = 10.0;

	shape = new btBoxShape(btVector3(0.5, 1.0, 0.2));
	trans.setOrigin(btVector3(x, 1.0, z));
	trans.setRotation(btQuaternion(btScalar(angle), btScalar(0), btScalar(0)));
	btDefaultMotionState* myMotionState = new btDefaultMotionState(trans);
	shape->calculateLocalInertia(mass, localInertia);
//	btRigidBody* body = new btRigidBody(mass, myMotionState, shape, localInertia, 1, 1, 0.3, 0);
	btRigidBody::btRigidBodyConstructionInfo constructionInfo(mass, myMotionState, shape, localInertia);
	constructionInfo.m_friction = 0.3;
	constructionInfo.m_restitution = 0;
	constructionInfo.m_linearSleepingThreshold = 2;
	constructionInfo.m_angularSleepingThreshold = 2;
	btRigidBody* body = new btRigidBody(constructionInfo);
	//body->setSleepingThresholds(2, 2);
	body->setDamping(0.6, 0.);
	blocks[id] = body;
	world->addRigidBody(body);

	return 0;
}

int get_block_matrix(int id, btScalar m[16]) {
	btRigidBody* body = blocks[id];
	btDefaultMotionState* myMotionState = (btDefaultMotionState*)body->getMotionState();
	myMotionState->m_graphicsWorldTrans.getOpenGLMatrix(m);

/*
	if (body->getActivationState() == 2) {
		glColor3d(0.5, 0.5, 0.5);
	} else if (body->getActivationState() == 3) {
		glColor3d(0.0, 1.0, 0.0);
	} else {
		glColor3d(1.0, 0.0, 0.0);
	}
*/

	glColor3d(1.0, 0.0, 0.0);

	glMultMatrixf(m);

	return 0;
}

int apply_force(int id) {
	btRigidBody* body = blocks[id];
	body->applyImpulse(btVector3(0, 0, -10), btVector3(0, 1, 0));
}

}
